using System;
using System.Runtime.InteropServices;

namespace DOOM.Music {
    class MusPlayer {
        private MidiDevice Device = null;
        private byte[] MusFile = null;


        [StructLayout(LayoutKind.Sequential)]
        private struct TIMECAPS {
            internal int wPeriodMin;
            internal int wPeriodMax;
        }

        [DllImport("winmm.dll")]
        private static extern int timeGetDevCaps(ref TIMECAPS lpTimeCaps, int uSize);
        [DllImport("winmm.dll")]
        private static extern int timeBeginPeriod(uint uPeriod);
        [DllImport("winmm.dll")]
        private static extern int timeEndPeriod(uint uPeriod);
        [DllImport("winmm.dll")]
        private static extern int timeSetEvent(uint uDelay, uint uResolution, TimerCallback lpFunction, int dwUser, int uFlags);
        private delegate void TimerCallback(uint wTimerId, uint msg, int dwUser, int dw1, int dw2);

        private const int TIMERR_NOERROR = 0;
        private const int TIME_ONESHOT = 0;
        private const int TIME_PERIODIC = 1;


        private uint TimerResolution;

        public readonly int ScoreLength = -1;
        public readonly int ScoreStart = -1;
        public readonly int PrimaryChannelCount = -1;
        public readonly int SecondaryChannelCount = -1;
        public readonly int InstrumentCount = -1;
        public readonly int[] Instruments = null;

        private int[] ChannelVolumes = new int[] { 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127 };

        private int ScorePointer = -1;

        TimerCallback MidiCallback;

        public MusPlayer(MidiDevice Device, byte[] MusFile) {
            this.Device = Device;
            this.MusFile = MusFile;
            if (MusFile.Length < 16) throw new Exception("Invalid MUS file");
            if (MusFile[0] != 'M' || MusFile[1] != 'U' || MusFile[2] != 'S' || MusFile[3] != 0x1A) throw new Exception("File does not have MUS header");

            this.ScoreLength = MusFile[4] + (MusFile[5] << 8);
            this.ScoreStart = MusFile[6] + (MusFile[7] << 8);
            this.PrimaryChannelCount = MusFile[8] + (MusFile[9] << 8);
            this.SecondaryChannelCount = MusFile[10] + (MusFile[11] << 8);
            this.InstrumentCount = MusFile[12] + (MusFile[13] << 8);
            this.Instruments = new int[InstrumentCount];

            for (int i = 0; i < this.InstrumentCount; ++i) {
                this.Instruments[i] = MusFile[i * 2 + 16] + (MusFile[i * 2 + 17] << 8);
            }

            this.ScorePointer = ScoreStart;

            TIMECAPS TimeCaps = new TIMECAPS();
            if (timeGetDevCaps(ref TimeCaps, Marshal.SizeOf(TimeCaps)) != TIMERR_NOERROR) throw new Exception("Could not get timer resolution");
            TimerResolution = (uint)Math.Min(Math.Max(TimeCaps.wPeriodMin, 1), TimeCaps.wPeriodMax);
            timeBeginPeriod(TimerResolution);

            MidiCallback =  new TimerCallback(SendMidi);
            SendMidi(0, 0, 0, 0, 0);
            
        }


        private void SendMidi(uint wTimerId, uint msg, int dwUser, int dw1, int dw2) {
            bool IsLast = false;

            while (!IsLast) {
                byte EventDescriptor = FetchByte();
                EventType Event = (EventType)(EventDescriptor & 0x70);
                int ChannelNumber = EventDescriptor & 0x0F;
                IsLast = (EventDescriptor & 0x80) != 0;

                if (ChannelNumber == 15) {
                    ChannelNumber = 9;
                } else if (ChannelNumber == 9) {
                    ChannelNumber = 15;
                }

                int Key;
                int Velocity;
                int ControllerNumber;
                int ControllerValue;
                int MidiControl;

                switch (Event) {
                    case EventType.ReleaseKey:
                        Key = FetchByte() & 0x7F;
                        this.Device.SendMidiMessage(MidiDevice.MidiMessage.KeyRelease, ChannelNumber, Key);
                        break;
                    case EventType.PressKey:
                        Key = FetchByte();
                        if ((Key & 0x80) != 0) {
                            ChannelVolumes[ChannelNumber] = FetchByte() & 0x7F;
                        }
                        Velocity = ChannelVolumes[ChannelNumber];
                        this.Device.SendMidiMessage(MidiDevice.MidiMessage.KeyPress, ChannelNumber, Key & 0x7F, Velocity);
                        break;
                    case EventType.PitchWheel:
                        Key = FetchByte() * 64;
                        Device.SendMidiMessage(MidiDevice.MidiMessage.PitchWheelChange, ChannelNumber, Key & 0x7F, (Key >> 7) & 0x7F);
                        break;
                    case EventType.SystemEvent:
                        MidiControl = 0x00;
                        int SystemEvent = FetchByte() & 0x7F;
                        switch (SystemEvent) {
                            case 10: MidiControl = 0x78; break;
                            case 11: MidiControl = 0x7B; break;
                            case 12: MidiControl = 0x7E; break;
                            case 13: MidiControl = 0x7F; break;
                            case 14: MidiControl = 0x79; break;
                        }
                        this.Device.SendMidiMessage(MidiDevice.MidiMessage.ControlChange, ChannelNumber, MidiControl);
                        break;
                    case EventType.ChangeController:
                        ControllerNumber = FetchByte() & 0x7F;
                        ControllerValue = FetchByte();// &0x7F;
                        switch (ControllerNumber) {
                            case 0:
                                this.Device.SendMidiMessage(MidiDevice.MidiMessage.PatchChange, ChannelNumber, ControllerValue);
                                break;
                            default:
                                MidiControl = 0x00;
                                switch (ControllerNumber) {
                                    case 1: MidiControl = 0x20; break;
                                    case 2: MidiControl = 0x01; break;
                                    case 3: MidiControl = 0x07; break;
                                    case 4: MidiControl = 0x0A; break;
                                    case 5: MidiControl = 0x0B; break;
                                    case 6: MidiControl = 0x5B; break;
                                    case 7: MidiControl = 0x5D; break;
                                    case 8: MidiControl = 0x40; break;
                                    case 9: MidiControl = 0x43; break;
                                }
                                this.Device.SendMidiMessage(MidiDevice.MidiMessage.ControlChange, ChannelNumber, MidiControl, ControllerValue);
                                break;
                        }
                        break;
                    case EventType.ScoreEnd:
                        this.ScorePointer = this.ScoreStart;
                        break;
                }
            }
            int NextDelay = 0;
            for (; ; ) {
                byte NextVal = FetchByte();
                NextDelay = NextDelay * 128 + (NextVal & 127);
                if ((NextVal & 0x80) == 0) break;
            }
            timeSetEvent((uint)((double)NextDelay * 1000.0d / 140.0d), TimerResolution, MidiCallback, 0, TIME_ONESHOT);
        }


        private enum EventType {
            ReleaseKey = 0 << 4,
            PressKey = 1 << 4,
            PitchWheel = 2 << 4,
            SystemEvent = 3 << 4,
            ChangeController = 4 << 4,
            ScoreEnd = 6 << 5
        };

        private byte FetchByte() {
            byte ToReturn = this.MusFile[this.ScorePointer++];
            if (this.ScorePointer >= this.ScoreLength + this.ScoreStart) {
                this.ScorePointer = this.ScoreStart;
            }
            return ToReturn;
        }
    }
}