using System;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.Collections;
using System.IO;

using DOOM.WAD;
using DOOM.Levels;

namespace DOOM {
    public partial class DOOM {

        private class SectorComponent {
            public VertexBuffer Buffer;
            public int Offset;
            public int Length;
            public CustomVertex.PositionColoredTextured[] Vertices;
            public SectorComponent(VertexBuffer Buffer, int Offset, int Length, CustomVertex.PositionColoredTextured[] Vertices) {
                this.Buffer = Buffer;
                this.Offset = Offset;
                this.Length = Length;
                this.Vertices = Vertices;
            }
        }
        private Dictionary<Wad.Sector, List<SectorComponent>> SectorToVerticesLinkup = new Dictionary<Wad.Sector, List<SectorComponent>>();
        private class TextureAnimation {
            public int GeometryCount = 0;
            public VertexBuffer Geometry = null;
            public List<Texture> TextureList = new List<Texture>();
            public int TextureIndex = 0;
            public int[] ScrollingWallOffsets = null;
        }
        
        private List<TextureAnimation> SolidWalls = null;
        private List<TextureAnimation> TransparentWalls = null;

        bool AddTextureToAnimation(ref TextureAnimation ToAddTo, Wad.WallTexture Texture) {
            return AddTextureToAnimation(ref ToAddTo, Texture.Texture);
        }
        bool AddTextureToAnimation(ref TextureAnimation ToAddTo, Bitmap Texture) {

            MemoryStream M = new MemoryStream();
            Bitmap TextureToUse = Texture;
            bool IsTransparent = false;
            int W = Texture.Width;
            int H = Texture.Height;
            unsafe {
                BitmapData BD = TextureToUse.LockBits(new Rectangle(0, 0, W, H), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                int* p = (int*)BD.Scan0.ToPointer();
                for (int j = 0; j < W * H; ++j) {
                    if (((*p++) & 0xFF000000) != 0xFF000000) {
                        IsTransparent = true;
                        break;
                    }
                }
                TextureToUse.UnlockBits(BD);
            }
            TextureToUse.Save(M, ImageFormat.Png);
            M.Seek(0, SeekOrigin.Begin);
            ToAddTo.TextureList.Add(TextureLoader.FromStream(RenderDevice, M));
            return IsTransparent;
        }


        public void InitialiseWallBuffers() {

            SolidWalls = new List<TextureAnimation>();
            TransparentWalls = new List<TextureAnimation>();

            for (int i = 0; i < CurrentLevel.Walls.Length; ++i) {
                bool IsTransparent = false;

                TextureAnimation CurrentWallBlock = new TextureAnimation();

                Wad.AnimatedTexture AnimatedTextureDetails = null;
                foreach (Wad.AnimatedTexture AT in CurrentLevel.WadFile.AnimatedTextures) {
                    if (AT.AllTextureNames.Contains(CurrentLevel.Walls[i].Texture.Name)) {
                        AnimatedTextureDetails = AT;
                        break;
                    }
                }

                Point TextureSize = new Point(CurrentLevel.Walls[i].Texture.Width, CurrentLevel.Walls[i].Texture.Height);

                if (AnimatedTextureDetails == null) {
                    // Single texture
                    IsTransparent = AddTextureToAnimation(ref CurrentWallBlock, CurrentLevel.Walls[i].Texture);
                } else {
                    foreach (Wad.WallTexture L in AnimatedTextureDetails.AllTextures) {
                        IsTransparent |= AddTextureToAnimation(ref CurrentWallBlock, L);
                    }
                    for (int j = 0; j < AnimatedTextureDetails.AllTextureNames.Count; ++j) {
                        if (AnimatedTextureDetails.AllTextureNames[j] == CurrentLevel.Walls[i].Texture.Name) {
                            CurrentWallBlock.TextureIndex = j;
                            break;
                        }
                    }
                }

                Dictionary<Wad.Sector, List<Level.Wall>> SortBySector = new Dictionary<Wad.Sector, List<Level.Wall>>();
                foreach (Level.Wall W in CurrentLevel.Walls[i].Walls) {
                    List<Level.Wall> SectorWallList;
                    if (!SortBySector.TryGetValue(W.Sector, out SectorWallList)) {
                        SortBySector[W.Sector] = new List<Level.Wall>();
                        SectorWallList = SortBySector[W.Sector];
                    }
                    SectorWallList.Add(W);
                }

                CurrentWallBlock.Geometry = new VertexBuffer(typeof(CustomVertex.PositionColoredTextured), 6 * CurrentLevel.Walls[i].Walls.Length, RenderDevice, Usage.None, CustomVertex.PositionColoredTextured.Format, Pool.Default);


                CustomVertex.PositionColoredTextured[] Vertices = new CustomVertex.PositionColoredTextured[6 * CurrentLevel.Walls[i].Walls.Length];
                int VertexIndex = 0;
                List<int> ScrollingWallOffsets = new List<int>();
                foreach (KeyValuePair<Wad.Sector, List<Level.Wall>> KVP in SortBySector) {

                    List<SectorComponent> SectorComponentToAddTo;
                    if (!SectorToVerticesLinkup.TryGetValue(KVP.Key, out SectorComponentToAddTo)) {
                        SectorComponentToAddTo = new List<SectorComponent>();
                        SectorToVerticesLinkup[KVP.Key] = SectorComponentToAddTo;
                    }

                    int InitialVertexIndex = VertexIndex;
                    foreach (Level.Wall W in KVP.Value) {

                        // Calculate texture coordinates

                        float WallLength = (float)Math.Sqrt((double)((W.Start.X - W.End.X) * (W.Start.X - W.End.X) + (W.Start.Y - W.End.Y) * (W.Start.Y - W.End.Y)));
                        float WallHeight = (float)(W.Ceiling - W.Floor);

                        float TuStart = (float)W.TextureOffset.X / (float)TextureSize.X;
                        float TuEnd = TuStart + WallLength / (float)TextureSize.X;
                        float TvStart = (float)W.TextureOffset.Y / (float)TextureSize.Y;
                        float TvEnd = TvStart + WallHeight / (float)TextureSize.Y;

                        if (W.Ceiling < W.Floor) {
                            Console.WriteLine("Wall " + W + " appears to be upside down.");
                        }

                        float CeilingHeight = W.Ceiling;
                        float FloorHeight = W.Floor;

                        if ((W.Flags & Wad.Line.LineFlags.TwoSided) != 0 && W.Type == Level.Wall.WallType.Middle) {
                            //if (W.Type == Level.Wall.WallType.Middle && IsTransparent) {

                            //if (CeilingHeight - FloorHeight > TextureSize.Y) {
                            if ((W.Flags & Wad.Line.LineFlags.LowerUnpegged) == 0) {
                                FloorHeight = CeilingHeight - TextureSize.Y;
                            } else {
                                CeilingHeight = FloorHeight + TextureSize.Y;
                            }
                            TvStart = 0.0f;
                            TvEnd = 1.0f;
                            //}
                        }

                        // Top-left triangle

                        Vertices[VertexIndex + 0].X = W.Start.X;
                        Vertices[VertexIndex + 0].Y = W.Start.Y;
                        Vertices[VertexIndex + 0].Z = CeilingHeight;
                        Vertices[VertexIndex + 0].Tu = TuStart;
                        Vertices[VertexIndex + 0].Tv = TvStart;

                        Vertices[VertexIndex + 1].X = W.End.X;
                        Vertices[VertexIndex + 1].Y = W.End.Y;
                        Vertices[VertexIndex + 1].Z = CeilingHeight;
                        Vertices[VertexIndex + 1].Tu = TuEnd;
                        Vertices[VertexIndex + 1].Tv = TvStart;

                        Vertices[VertexIndex + 2].X = W.Start.X;
                        Vertices[VertexIndex + 2].Y = W.Start.Y;
                        Vertices[VertexIndex + 2].Z = FloorHeight;
                        Vertices[VertexIndex + 2].Tu = TuStart;
                        Vertices[VertexIndex + 2].Tv = TvEnd;

                        // Bottom-right triangle

                        Vertices[VertexIndex + 3].X = W.Start.X;
                        Vertices[VertexIndex + 3].Y = W.Start.Y;
                        Vertices[VertexIndex + 3].Z = FloorHeight;
                        Vertices[VertexIndex + 3].Tu = TuStart;
                        Vertices[VertexIndex + 3].Tv = TvEnd;

                        Vertices[VertexIndex + 4].X = W.End.X;
                        Vertices[VertexIndex + 4].Y = W.End.Y;
                        Vertices[VertexIndex + 4].Z = CeilingHeight;
                        Vertices[VertexIndex + 4].Tu = TuEnd;
                        Vertices[VertexIndex + 4].Tv = TvStart;

                        Vertices[VertexIndex + 5].X = W.End.X;
                        Vertices[VertexIndex + 5].Y = W.End.Y;
                        Vertices[VertexIndex + 5].Z = FloorHeight;
                        Vertices[VertexIndex + 5].Tu = TuEnd;
                        Vertices[VertexIndex + 5].Tv = TvEnd;

                        // Set colour (brightness)

                        for (int j = 0; j < 6; ++j) {
                            Vertices[VertexIndex + j].Color = W.LightLevel + (W.LightLevel << 8) + (W.LightLevel << 16);
                            /*if (W.Type == Level.Wall.WallType.Middle) {
                                Vertices[VertexIndex + j].Color &= 0x0000FF;
                            } else if (W.Type == Level.Wall.WallType.Lower) {
                                Vertices[VertexIndex + j].Color &= 0xFF0000;
                            } else {
                                Vertices[VertexIndex + j].Color &= 0x00FF00;
                            }*/
                        }

                        // Move on...

                        if (W.Scrolling) {
                            //ScrollingWall S = new ScrollingWall(
                            ScrollingWallOffsets.Add(VertexIndex);
                        }

                        VertexIndex += 6;
                    }
                    int NumberOfVertices = VertexIndex - InitialVertexIndex;
                    int Size = NumberOfVertices * CustomVertex.PositionColoredTextured.StrideSize;
                    CustomVertex.PositionColoredTextured[] SectorComponentVertices = new CustomVertex.PositionColoredTextured[NumberOfVertices];
                    for (int j = 0; j < NumberOfVertices; ++j) {
                        SectorComponentVertices[j] = Vertices[j + InitialVertexIndex];
                    }
                    SectorComponentToAddTo.Add(new SectorComponent(CurrentWallBlock.Geometry, InitialVertexIndex * CustomVertex.PositionColoredTextured.StrideSize, Size, SectorComponentVertices));

                }  
            
                

                GraphicsStream Writer = CurrentWallBlock.Geometry.Lock(0, 0, 0);
                Writer.Write(Vertices);
                CurrentWallBlock.Geometry.Unlock();
                CurrentWallBlock.GeometryCount = VertexIndex / 3;
                CurrentWallBlock.ScrollingWallOffsets = ScrollingWallOffsets.ToArray();

                if (IsTransparent) {
                    TransparentWalls.Add(CurrentWallBlock);
                } else {
                    SolidWalls.Add(CurrentWallBlock);
                }
            }

        }

        public class DrawableSubsector {
            public int WallTexture;
            public VertexBuffer Walls;
        }



        public void DrawWalls() {



            RenderDevice.TextureState[0].ColorOperation = TextureOperation.Modulate2X;
            RenderDevice.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
            RenderDevice.TextureState[0].ColorArgument2 = TextureArgument.Diffuse;

            RenderDevice.VertexFormat = CustomVertex.PositionColoredTextured.Format;
            foreach (TextureAnimation T in SolidWalls) {
                RenderDevice.SetTexture(0, T.TextureList[T.TextureIndex]);
                RenderDevice.SetStreamSource(0, T.Geometry, 0);
                RenderDevice.DrawPrimitives(PrimitiveType.TriangleList, 0, T.GeometryCount);
            }

            RenderDevice.RenderState.ZBufferWriteEnable = false;
            for (int j = 0; j < 2; j++) {
                foreach (TextureAnimation T in TransparentWalls) {
                    RenderDevice.SetTexture(0, T.TextureList[T.TextureIndex]);
                    RenderDevice.SetStreamSource(0, T.Geometry, 0);
                    RenderDevice.DrawPrimitives(PrimitiveType.TriangleList, 0, T.GeometryCount);
                }
                RenderDevice.RenderState.ZBufferWriteEnable = true;    
            }
        }        
    }
}
