// Sector decoding
using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections.Generic;

namespace DOOM.WAD {
    public partial class Wad {

        public static Sector[] DecodeSectors(Lump SectorFile) {
            Sector[] Return = new Sector[SectorFile.Size / 26];
            for (int i = 0; i < SectorFile.Size; i += 26) {
                object FindFloor;
                object FindCeiling;

                string FloorName = "";
                string CeilingName = "";

                for (int j = 0; j < 8; ++j) {
                    FloorName += (char)SectorFile.Data[i + j + 4];
                    CeilingName += (char)SectorFile.Data[i + j + 12];
                    
                }
                FloorName = FixString(FloorName);
                CeilingName = FixString(CeilingName);

                bool SkyCeiling = (CeilingName == "F_SKY1");
                bool SkyFloor = (FloorName == "F_SKY1");
                
                FindFloor = SectorFile.Parent.LumpIndex[FloorName];
                FindCeiling = SectorFile.Parent.LumpIndex[CeilingName];               
                

                if (FindFloor == null) throw new Exception("Floor " + CeilingName + " not found.");
                if (FindCeiling == null) throw new Exception("Ceiling " + CeilingName + " not found.");

                FindFloor = SectorFile.Parent.Lumps[(int)FindFloor];
                FindCeiling = SectorFile.Parent.Lumps[(int)FindCeiling];

                Return[i / 26] = new Sector(
                    i / 26,
                    FixSignedShort(SectorFile.Data[i+0] + (SectorFile.Data[i+1] << 8)),
                    FixSignedShort(SectorFile.Data[i+2] + (SectorFile.Data[i+3] << 8)),
                    (Lump)FindFloor, (Lump)FindCeiling, 
                    SectorFile.Data[i+20] + (SectorFile.Data[i+21] << 16),
                    (Sector.SectorType)(SectorFile.Data[i+22] + (SectorFile.Data[i+23] << 16)),
                    SectorFile.Data[i+24] + (SectorFile.Data[i+25] << 16),
                    SkyCeiling, SkyFloor,
                    CeilingName, FloorName
               );
            }
            return Return;
        }
        public class Sector {

            /// <summary>
            /// Indetifier
            /// </summary>
            public readonly int ID;

            /// <summary>
            /// The level of the floor
            /// </summary>
            public readonly int FloorHeight;
            /// <summary>
            /// The level of the ceiling
            /// </summary>
            public readonly int CeilingHeight;

            /// <summary>
            /// Floor texture
            /// </summary>
            public readonly Lump FloorTexture;
            /// <summary>
            /// Ceiling texture
            /// </summary>
            public readonly Lump CeilingTexture;

            /// <summary>
            /// Light level between 0 (darkest) and 255 (brightest)
            /// </summary>
            public readonly int LightLevel;

            /// <summary>
            /// Special sector type
            /// </summary>
            public readonly SectorType Type;

            /// <summary>
            /// Tag used by linedefs
            /// </summary>
            public readonly int Tag;

            public readonly bool SkyCeiling;
            public readonly bool SkyFloor;

            public readonly string SkyTextureName;
            public readonly string FloorTextureName;

            public List<Sector> AdjacentSectors = null;




            public enum SectorType {
                /// <summary>
                /// No special characteristics
                /// </summary>
                Normal = 0x00,
                /// <summary>
                /// Light blinks off randomly
                /// </summary>
                LightRandomOff = 0x01,
                /// <summary>
                /// Light blinks every half-second
                /// </summary>
                LightBlinkHalfSecond = 0x02,
                /// <summary>
                /// Light blinks every second
                /// </summary>
                LightBlinkWholeSecond = 0x03,
                /// <summary>
                /// Light blinks every half second, -20% health damage
                /// </summary>
                LightBlinkAndDamage20 = 0x04,
                /// <summary>
                /// -10% health damage
                /// </summary>
                Damage10 = 0x05,
                /// <summary>
                /// -5% health damage
                /// </summary>
                Damage5 = 0x07,
                /// <summary>
                /// Light oscillates up and down
                /// </summary>
                LightOscillates = 0x08,
                /// <summary>
                /// Sector is flagged as a secret area
                /// </summary>
                Secret = 0x09,
                /// <summary>
                /// Door closes 30 seconds after the level starts
                /// </summary>
                DoorClose = 0x0A,
                /// <summary>
                /// -20% health damage, when it becomes less than 11% the level ends
                /// </summary>
                End = 0x0B,
                /// <summary>
                /// Light blinks every half-second (synchronized)
                /// </summary>
                LightBlinkHalfSecondSynchronized = 0x0D,
                /// <summary>
                /// Light blinks every second (synchronized)
                /// </summary>
                LightBlinkWholeSecondSynchronized = 0x0C,
                /// <summary>
                /// Door opens 30 seconds after the level starts
                /// </summary>
                DoorOpen = 0x0E,
                /// <summary>
                /// -20% health damage
                /// </summary>
                Damage20 = 0x10,
                /// <summary>
                /// Light flickers randomly
                /// </summary>
                LightFlicker = 0x11,
                /// <summary>
                /// Crushing ceiling
                /// </summary>
                CrushingCeiling = 0x06,
                /// <summary>
                /// Ammo creator
                /// </summary>
                AmmoCreator = 0x0F

            };

            public Sector(int ID, int FloorHeight, int CeilingHeight, Lump FloorTexture, Lump CeilingTexture, int LightLevel, SectorType Type, int Tag, bool SkyCeiling, bool SkyFloor, string CeilingTextureName, string FloorTextureName) {
                this.ID = ID;
                this.FloorHeight = FloorHeight;
                this.CeilingHeight = CeilingHeight;
                this.FloorTexture = FloorTexture;
                this.CeilingTexture = CeilingTexture;
                this.LightLevel = LightLevel;
                this.Type = Type;
                this.Tag = Tag;
                this.SkyCeiling = SkyCeiling;
                this.SkyFloor = SkyFloor;
                this.SkyTextureName = CeilingTextureName;
                this.FloorTextureName = FloorTextureName;
            }

            public List<Segment> ContainedSegments;
            public List<Line> ContainedLineDefs;
            public void IndexLineDefs(Line[] AllLineDefs, Side[] AllSideDefs, Segment[] AllSegments) {
                ContainedLineDefs = new List<Line>();
                
                foreach (Side S in AllSideDefs) {
                    if (S.Sector == this.ID) {
                        // S is a valid side def...
                        foreach (Line L in AllLineDefs) {
                            if ((L.RightSide == S.ID || L.LeftSide == S.ID)) {
                                
                                if (L.RightSide == S.ID) L.RightSector = this.ID;
                                if (L.LeftSide == S.ID) L.LeftSector = this.ID;

                                if (!ContainedLineDefs.Contains(L)) ContainedLineDefs.Add(L);
                                //L.Sector = this.ID;
                            }
                        }
                    }
                }

                this.ContainedSegments = new List<Segment>();
                foreach (Segment S in AllSegments) {
                    if (this.ContainedLineDefs.Contains(AllLineDefs[S.Line])) {
                        Wad.Line FoundLine = AllLineDefs[S.Line];
                        if (FoundLine.RightSector == this.ID) this.ContainedSegments.Add(S);
                    }
                }

            }
        }
    }
}