// Basic WAD loading/directory reading code.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Drawing;

namespace DOOM.WAD {
    /// <summary>
    /// WAD file reading class.
    /// </summary>
    public partial class Wad {

        /// <summary>
        /// Indentifies the type of a lump.
        /// </summary>
        public enum LumpType {
            Unknown, Palette, ColourMap, DosExit, Demo, Texture, PatchIndex, MidiMap, GusMap, PcSpeakerSound, Wav, Song, Flat, Graphic, Things, LineDefinitions, SideDefinitions, VertexList, SegmentList, SubsectorList, NodeList, SectorList, RejectLump, Blockmap, Marker
        };

        public readonly Hashtable LumpIndex;

        /// <summary>
        /// The total number of lumps.
        /// </summary>
        public readonly int LumpCount;

        /// <summary>
        /// The list of lumps inside the WAD file.
        /// </summary>
        public readonly List<Lump> Lumps;        

        /// <summary>
        /// Offset to the WAD's directory information.
        /// </summary>
        private readonly int DirectoryOffset;
        
        /// <summary>
        /// Load and read a WAD file
        /// </summary>
        /// <param name="Filename"></param>
        public Wad(string Filename) {
            using (BinaryReader WadReader = new BinaryReader(new FileStream(Filename, FileMode.Open))) {
                
                // Check the WAD header
                string Header = new string(WadReader.ReadChars(4));
                if (Header != "IWAD") { // Also support PWAD..?
                    throw new Exception("Invalid WAD header.");
                }

                // Get the number of lumps and the offset to the WAD's directory information.
                this.LumpCount = WadReader.ReadInt32();
                this.DirectoryOffset = WadReader.ReadInt32();

                // Reset the list of lumps.
                this.Lumps = new List<Lump>(LumpCount);
                this.LumpIndex = new Hashtable(LumpCount);

                // Read all the lumps...

                string LumpFolder = "";
                for (int i = 0; i < LumpCount; ++i) {
                    
                    // Seek to the directory entry for the current lump
                    WadReader.BaseStream.Seek(DirectoryOffset + i * 16, SeekOrigin.Begin);
                    
                    // Read the offset, size and name information from the directory.
                    int LumpOffset = WadReader.ReadInt32();
                    int LumpSize = WadReader.ReadInt32();
                    
                    // Get the lump name, stripping those pesky NULs off the end.
                    string LumpName = FixString(new string(WadReader.ReadChars(8)));

                    LumpIndex[LumpName] = i;

                    // Folder name
                    if (LumpSize == 0) LumpFolder = LumpName;

                    // Seek the data:
                    WadReader.BaseStream.Seek(LumpOffset, SeekOrigin.Begin);
                    
                    // Read the data in and add the new lump to the list.
                    Lump ToAdd = new Lump(this, LumpName, LumpFolder, WadReader.ReadBytes(LumpSize));
                    this.Lumps.Add(ToAdd);
                    
                    // If it's a palette, process it. We need to process palettes before
                    // all the other graphics, hence it is done here.
                    if (ToAdd.Name == "PLAYPAL") DecodePalette(ToAdd);

                }
                if (Palettes == null) throw new Exception("No palette information in the WAD file.");

                // Decode the patch names:
                object PatchNamesLump = LumpIndex["PNAMES"];
                if (PatchNamesLump == null) throw new Exception("No patch index file in the WAD file.");
                DecodePatchIndices(this.Lumps[(int)PatchNamesLump]);

                // Decode the texture names:
                Textures = new Hashtable();
                foreach (Lump L in Lumps) {
                    if (L.Type == LumpType.Texture) {
                        DecodeTextures(L);
                    }
                }

                // Hunt all the animated textures;
                WorkOutWhichAreAnimatedTextures();

            }
        }
        
        /// <summary>
        /// Represents a lump ('file') within the wad.
        /// </summary>
        public class Lump {
            
            /// <summary>
            /// Internal name of the lump.
            /// </summary>
            public readonly string Name;
            
            /// <summary>
            /// Size of the lump in bytes.
            /// </summary>
            public readonly int Size;
            
            /// <summary>
            /// The data of the lump.
            /// </summary>
            public readonly byte[] Data;

            /// <summary>
            /// The lump type.
            /// </summary>
            public readonly LumpType Type;

            /// <summary>
            /// A general-purpose object that can contain a decoded version of the data.
            /// </summary>
            public object DecodedData {
                get {
                    if (_DecodedData == null) {
                        switch (Type) {
                            case LumpType.Flat:
                                _DecodedData = Wad.DecodeFlat(this);
                                break;
                            case LumpType.Graphic:
                                _DecodedData = Wad.DecodeGraphic(this);
                                break;
                            case LumpType.VertexList:
                                _DecodedData = Wad.DecodeVertices(this);
                                break;
                            case LumpType.LineDefinitions:
                                _DecodedData = Wad.DecodeLine(this);
                                break;
                            case LumpType.SideDefinitions:
                                _DecodedData = Wad.DecodeSide(this);
                                break;
                            case LumpType.SegmentList:
                                _DecodedData = Wad.DecodeSegments(this);
                                break;
                            case LumpType.SectorList:
                                _DecodedData = Wad.DecodeSectors(this);
                                break;
                            case LumpType.SubsectorList:
                                _DecodedData = Wad.DecodeSubSectors(this);
                                break;
                            case LumpType.Things:
                                _DecodedData = Wad.DecodeThings(this);
                                break;
                            case LumpType.NodeList:
                                _DecodedData = Wad.DecodeNodes(this);
                                break;
                        }
                    }
                    return _DecodedData;
                }
            }
            private object _DecodedData = null;

            /// <summary>
            /// The "folder" the file is inside.
            /// </summary>
            public readonly string Folder;

            public readonly Wad Parent;

            public Lump(Wad Parent, string Name, string Folder, byte[] Data) {
                this.Parent = Parent;
                this.Name = Name;
                this.Folder = Folder;
                this.Data = Data;
                this.Size = Data.Length;
                this.Type = IndentifyLumpType();
            }

            /// <summary>
            /// Work out what type of lump this is.
            /// </summary>
            private LumpType IndentifyLumpType() {

                if (this.Size == 0) return LumpType.Marker;

                if (this.Folder.Length != 0 && this.Folder[0] == 'F') return LumpType.Flat;

                switch (this.Name) {
                    case "PLAYPAL" : return LumpType.Palette;
                    case "COLORMAP" : return  LumpType.ColourMap;
                    case "ENDOOM" : return LumpType.DosExit;
                    case "PNAMES" : return  LumpType.PatchIndex;
                    case "GENMIDI" : return  LumpType.MidiMap;
                    case "THINGS" : return  LumpType.Things;
                    case "LINEDEFS" : return  LumpType.LineDefinitions;
                    case "SIDEDEFS" : return LumpType.SideDefinitions;
                    case "VERTEXES" : return  LumpType.VertexList;
                    case "SEGS" : return LumpType.SegmentList;
                    case "SSECTORS" : return LumpType.SubsectorList;
                    case "NODES" : return LumpType.NodeList;
                    case "SECTORS" : return LumpType.SectorList;
                    case "REJECT" : return LumpType.RejectLump;
                    case "BLOCKMAP" : return LumpType.Blockmap;
                    case "DOOMERR": return LumpType.DosExit;
                    case "POINTS": return LumpType.VertexList;
                    case "LINES": return LumpType.LineDefinitions;
                }

                if (this.Name.StartsWith("D_")) return LumpType.Song;
                if (this.Name.StartsWith("DEMO")) return LumpType.Demo;
                if (this.Name.StartsWith("DP")) return LumpType.PcSpeakerSound;
                if (this.Name.StartsWith("DS")) return LumpType.Wav;
                if (this.Name.StartsWith("TEXTURE")) return LumpType.Texture;
                if (this.Name.StartsWith("DMXGUS")) return LumpType.GusMap;

                if (this.Data.Length < 8) return LumpType.Unknown;

                int W = this.Data[0] + this.Data[1] * 256;
                int H = this.Data[2] + this.Data[3] * 256;

                if (W > 1024 || H > 1024) return LumpType.Unknown;

                return LumpType.Graphic;
            }
        }

    }
}
